import sys
import os
import ctypes
import numpy as np
import time

from printer import print_line

from compiler   import *
from constructs import *
from utils import *

def call_pipe(app_data):
    rows = app_data['rows']
    cols = app_data['cols']

    img_data = app_data['img_data']
    IN = img_data['IN']
    OUT = img_data['OUT']
    alpha = app_data['alpha']
    beta = app_data['beta']

    # lib function name
    func_name = 'pipeline_'+app_data['app']
    pipe_func = app_data[func_name]

    # lib function args
    pipe_args = []
    pipe_args += [ctypes.c_int(cols)]
    pipe_args += [ctypes.c_int(rows)]
    pipe_args += [ctypes.c_float(alpha)]
    pipe_args += [ctypes.c_float(beta)]
    pipe_args += [ctypes.c_void_p(IN.ctypes.data)]
    pipe_args += [ctypes.c_void_p(OUT.ctypes.data)]

    # call lib function
    pipe_func(*pipe_args)
    
    return

def laplacian(app_data):

    it  = 0
    app_args = app_data['app_args']
   
    runs = int(app_args.runs)
    timer = app_args.timer
    if timer == True:
        t1 = time.time()

    avg = 0
    while it < runs :
        t1 = time.time()
        call_pipe(app_data)
        t2 = time.time()

        time_taken = float(t2) - float(t1)
        avg += time_taken
        print("")
        print("[exec_pipe] : time taken to execute = ", (time_taken * 1000), " ms")

        it += 1

    print ("average time ", avg/runs*1000, " ms")

    return avg/runs*1000

